(* File:    CBN_Games.sml
   Author:  John Longley
   Date:    July 2008 (started)
*)

(* SML implementation of Lamarche game model 
   with repetitive backtracking exponential. 
   Semantic infrastructure for Lingay runtime. *)

(*
fun CBN_Games() = use "CBN_Games.sml" ;
fun CBN_Games_Test() = use "CBN_Games_Test.sml" ;
*)

signature CBN_GAMES = sig

   eqtype ground  (* equality only used for testing! *)
   type label = string
   type index = int
   datatype omove = BasicO of ground | MainO of omove
	       | ProdO of label * omove | InjO of label * omove
	       | LeftO of pmove | RightO of omove | FoldO of omove
	       | TagO of index * omove 
   and pmove = BasicP of ground | MainP of pmove
	       | ProdP of label * pmove | InjP of label * pmove
	       | LeftP of omove | RightP of pmove | FoldP of pmove
	       | TagP of index * pmove 
   datatype ('s,'t) Mmove = Source of 's | Target of 't 
   datatype ('o,'p) strategy = strategy of 'o -> 'p * ('o,'p) strategy ;
   type morphism = ((pmove,omove) Mmove, (omove,pmove) Mmove) strategy ;
   type relabelling = {relabO : omove -> omove, relabP : pmove -> pmove} ;

   val blank    : ground
   val % : morphism * (pmove,omove) Mmove -> (omove,pmove) Mmove * morphism
   val RELABEL  : relabelling -> morphism
   val ID       : morphism
   val THEN     : morphism * morphism -> morphism  (* forward composition *)
   val PRODUCT  : (label * morphism) list -> morphism
   val PROJ     : (label list * label list) list -> morphism
   val CURRY    : label * label -> morphism -> morphism
   val UNCURRY  : label * label -> morphism -> morphism
   val ABSORB   : (ground -> morphism) -> morphism
   val SUPPLY   : ground -> morphism -> morphism
   val BANG     : morphism -> morphism
   val DERELICT : morphism
   val COMULT   : morphism
   val CONTRACT : morphism
   val LIFT     : (ground -> ground * morphism) -> morphism
   val UNIT     : ground -> morphism
   val FLATTEN  : morphism
   val STRENGTH : label -> (ground -> ground) -> morphism
   val FIX      : morphism
   val ZERO     : morphism

   val trace    : string -> morphism -> morphism
(* for debugging: *)
   val omoveToS : omove -> string
   val MomoveToS: (pmove,omove) Mmove -> string
end ;


functor CBN_Games (eqtype ground ; val blank : ground ; 
                   val gToS : ground -> string) 
	: CBN_GAMES where type ground = ground
	= struct

open Pervasives ;

type ground = ground
type label = string
type index = int ;

exception LabelError of label
exception IndexError of index ;

(* Moves and strategies for a game *)

datatype omove = BasicO of ground | MainO of omove
	       | ProdO of label * omove | InjO of label * omove
	       | LeftO of pmove | RightO of omove | FoldO of omove
	       | TagO of index * omove 
     and pmove = BasicP of ground | MainP of pmove
	       | ProdP of label * pmove | InjP of label * pmove
	       | LeftP of omove | RightP of pmove | FoldP of pmove
	       | TagP of index * pmove ;

datatype ('o,'p) strategy = strategy of 'o -> 'p * ('o,'p) strategy ;


(* Moves and strategies for morphisms *)

datatype ('s,'t) Mmove = Source of 's | Target of 't ;

type Momove = (pmove,omove) Mmove ;
type Mpmove = (omove,pmove) Mmove ;
datatype OPmove = MO of Momove | MP of Mpmove ;

type morphism = (Momove,Mpmove) strategy ;
val morphism  = (strategy : (Momove -> Mpmove * morphism) -> morphism) ;

infix %
fun (strategy f : morphism) % x = f x ;

val blank = blank ;

(* Outputting a trace of a strategy *)

fun omoveToS (BasicO x) = "BasicO " ^ gToS x
  | omoveToS (MainO om) = "MainO (" ^ omoveToS om ^ ")"
  | omoveToS (ProdO (l,om)) = "ProdO (" ^ l ^ ", " ^ omoveToS om ^ "))"
  | omoveToS (LeftO pm) = "LeftO (" ^ pmoveToS pm ^ ")"
  | omoveToS (RightO om) = "RightO (" ^ omoveToS om ^ ")"
  | omoveToS (TagO (i,om)) = "TagO (" ^ Int.toString i ^ ", " ^
			     omoveToS om ^ "))"
  | omoveToS _ = "?"
and pmoveToS (BasicP x) = "BasicP " ^ gToS x
  | pmoveToS (MainP pm) = "MainP (" ^ pmoveToS pm ^ ")"
  | pmoveToS (ProdP (l,pm)) = "ProdP (" ^ l ^ ", " ^ pmoveToS pm ^ "))"
  | pmoveToS (LeftP om) = "LeftP (" ^ omoveToS om ^ ")"
  | pmoveToS (RightP pm) = "RightP (" ^ pmoveToS pm ^ ")"
  | pmoveToS (TagP (i,pm)) = "TagP (" ^ Int.toString i ^ ", " ^
			     pmoveToS pm ^ "))"
  | pmoveToS _ = "?"

fun MomoveToS (Source pm) = "Source (" ^ pmoveToS pm ^ ")"
  | MomoveToS (Target om) = "Target (" ^ omoveToS om ^ ")"
fun MpmoveToS (Source om) = "Source (" ^ omoveToS om ^ ")"
  | MpmoveToS (Target pm) = "Target (" ^ pmoveToS pm ^ ")"


fun trace tr (strategy f) = strategy
   (fn om => (TextIO.print ("    "^tr^" O:  "^MomoveToS om^"\n") ;
              case f om of (pm,rest) =>
              (TextIO.print ("    "^tr^" P:  "^MpmoveToS pm^"\n") ;
	       (pm, trace tr rest)))) ;


(* Relabelling of target moves. *)

type relabelling = {relabO : omove -> omove, relabP : pmove -> pmove} ;

infix +> +<
exception ParityError ;

fun s +> (r as {relabO = ro, relabP = rp}) = strategy 
    (fn Target om => (case s % Target (ro om) of 
	   (Target pm',s') => (Target (rp pm'), s' +> r)
	 | (Source om',s') => (Source om', s' +< r))
      | _ => raise ParityError)
and s +< (r as {relabO = ro, relabP = rp}) = strategy 
    (fn Source pm => (case s % Source pm of 
	   (Target pm',s') => (Target (rp pm'), s' +> r)
	 | (Source om',s') => (Source om', s' +< r))
      | _ => raise ParityError) ;


(* Identity and composition of strategies etc. *)

local
fun idStrat' (Target y) = (Source y, strategy idStrat'')
  | idStrat' _ = raise ParityError
and idStrat'' (Source x) = (Target x, strategy idStrat')
  | idStrat'' _ = raise ParityError 
in
val ID = strategy idStrat'
end ;

fun RELABEL r = ID +> r ;

local
fun play2 s1 s2 m2 =
    (case s2 % m2 of
      (Target z,s2') => (Target z,s1,s2')
    | (Source y,s2') => play1 s1 s2' (Target y))
and play1 s1 s2 m1 =
    (case s1 % m1 of
      (Source x,s1') => (Source x,s1',s2)
    | (Target y,s1') => play2 s1' s2 (Source y))
in
fun compose s1 s2 = strategy
    (fn Target z => (case play2 s1 s2 (Target z) of
        (Target z',s1',s2') => (Target z', compose s1' s2')
      | (Source x',s1',s2') => (Source x', compose' s1' s2'))
     | _ => raise ParityError)
and compose' s1 s2 = strategy
    (fn Source x => (case play1 s1 s2 (Source x) of
        (Target z',s1',s2') => (Target z', compose s1' s2')
      | (Source x',s1',s2') => (Source x', compose' s1' s2'))
     | _ => raise ParityError)       
	(* can shorten stuff like this if we eschew parity checking *)
end ;

infix THEN ;
fun (s1 THEN s2) = compose s1 s2 ;  


(* Labelled products of strategies *)

type 'a lablist = (label * 'a) list ;

fun lookupLabel alist x = errorLookup LabelError alist x ;

fun productStrats complist = strategy
    (fn Target (ProdO (c,om)) =>
        (case lookupLabel complist c of s =>
         case s % Target om of 
	   (Target pm,s')  => (Target (ProdP (c,pm)), 
                               productStrats (updateAlist (c,s') complist))
	 | (Source om',s') => (Source (ProdO (c,om')),
       	                       productStrats' (updateAlist (c,s') complist)))
      | _ => raise ParityError) 
and productStrats' complist = strategy
    (fn Source (ProdP (c,pm)) =>
        (case lookupLabel complist c of s =>
         case s % Source pm of 
	   (Target pm,s')  => (Target (ProdP (c,pm)), 
	                       productStrats (updateAlist (c,s') complist))
	 | (Source om',s') => (Source (ProdO (c,om')),
       	                       productStrats' (updateAlist (c,s') complist)))
      | _ => raise ParityError) ;
val PRODUCT = productStrats ;

(* Generalized projection.
   If X and Y are (nested) product types, and f is an injection from
   components of Y to components of X, we get a projection morphism 
   from X to Y. This gives everything we get from the affine 
   symmetric monoidal structure *)

type comp = label list   (* innermost first *)
type compRenaming = (comp * comp) list ;
exception CompError of comp ;

fun StoTcomp renaming (c:comp) = optLookup renaming c ;
fun TtoScomp renaming (c:comp) = optLookup' renaming c ;

fun addLabels Prod [] om = om
  | addLabels Prod (l::rest) om = addLabels Prod rest (Prod (l,om)) ;

fun relabO' renaming comp om =
    (case TtoScomp renaming comp of
        SOME comp' => addLabels ProdO comp' om
      | NONE => (case om of
             ProdO (l,om') => relabO' renaming (l::comp) om'
           | _ => raise CompError comp)) ;

fun relabP' renaming comp om =
    (case StoTcomp renaming comp of
        SOME comp' => addLabels ProdP comp' om
      | NONE => (case om of
             ProdP (l,om') => relabP' renaming (l::comp) om'
           | _ => raise CompError comp)) ;

fun PROJ renaming =
    RELABEL {relabO = relabO' renaming [],
             relabP = relabP' renaming []} ;


(* Closed structure: currying and uncurrying *)

exception CurryError of OPmove ;
exception UncurryError of OPmove ;

fun curryPmove (l1,l2) (Target pm') = 
     Target (RightP pm')
  | curryPmove (l1,l2) (pm as Source (ProdO (l,om))) =
     if l=l1 then Source om
     else if l=l2 then Target (LeftP om)
     else raise CurryError (MP pm)
  | curryPmove _ pm = raise CurryError (MP pm) ;

fun curryOmove (l1,l2) (Target om') = 
     Target (RightO om')
  | curryOmove (l1,l2) (om as Source (ProdP (l,pm))) =
     if l=l1 then Source pm
     else if l=l2 then Target (LeftO pm)
     else raise CurryError (MO om)
  | curryOmove _ om = raise CurryError (MO om) ;
  
fun uncurryPmove (l1,l2) (Target (RightP pm')) = 
     Target pm'
  | uncurryPmove (l1,l2) (Target (LeftP om)) = 
     Source (ProdO (l2,om))
  | uncurryPmove (l1,l2) (Source om) =
     Source (ProdO (l1,om))
  | uncurryPmove _ pm = raise UncurryError (MP pm) ;

fun uncurryOmove (l1,l2) (Target (RightO om')) = 
     Target om'
  | uncurryOmove (l1,l2) (Target (LeftO pm)) = 
     Source (ProdP (l2,pm))
  | uncurryOmove (l1,l2) (Source pm) =
     Source (ProdP (l1,pm))
  | uncurryOmove _ om = raise UncurryError (MO om) ;

local fun curry' ls s om = 
   case s % uncurryOmove ls om of 
     (pm,s') => (curryPmove ls pm, strategy (curry' ls s'))
in fun CURRY ls s = strategy (curry' ls s)
end ;

local fun uncurry' ls s om =
   case s % curryOmove ls om of 
     (pm,s') => (uncurryPmove ls pm, strategy (uncurry' ls s'))
in fun UNCURRY ls s = strategy (uncurry' ls s)
end ;

exception AbsorbError ;
exception SupplyError ;

fun extractGround (BasicO x) = (x, BasicO blank)
  | extractGround (RightO om) =
      let val (x,om') = extractGround om in (x, RightO om') end
  | extractGround (ProdO (l,om)) =
      let val (x,om') = extractGround om in (x, ProdO (l,om')) end
  | extractGround (TagO (i,om)) =
      let val (x,om') = extractGround om in (x, TagO (i,om')) end
  | extractGround (FoldO om) =
      let val (x,om') = extractGround om in (x, FoldO om') end
  | extractGround _ = raise AbsorbError ;

fun ABSORB G = strategy
  (fn Target om => 
      let val (x,om') = extractGround om in G x % Target om end
    | _ => raise AbsorbError) ;

fun SUPPLY x s = strategy
  (fn Target (BasicO _) => s % (Target (BasicO x))
    | _ => raise SupplyError) ;


(* Bang structure *)

(* Action of bang on morphisms *)

local

type ('a,'b) bangAdmin = 
     {targetHistory : (('a,'b) strategy * index) list,
      lastSourceTag : index,
      lastTargetTag : index}

val initAdmin = 
    {targetHistory = [], lastSourceTag = 0, lastTargetTag = 0} ;

fun lookupStrat i (admin : ('a,'b) bangAdmin) = 
    List.nth (#targetHistory admin, #lastTargetTag admin - i) ;

fun incSourceTag (admin : ('a,'b) bangAdmin) =
    {targetHistory = #targetHistory admin,
     lastSourceTag = #lastSourceTag admin + 1,
     lastTargetTag = #lastTargetTag admin} ;

fun updateAdmin (s,tag,inc) (admin : ('a,'b) bangAdmin) =
   {targetHistory = (s,tag) :: #targetHistory admin,
    lastSourceTag = #lastSourceTag admin + inc,
    lastTargetTag = #lastTargetTag admin + 1}  ;

exception BangError ;

fun bangStrat' s0 admin = strategy
    (fn Target (TagO (i,om)) =>
	let val (s,lastTag) = if i=0 then (s0,0) else lookupStrat i admin
        in case s % Target om of
           (Target pm, s') => 
	     (Target (TagP (#lastTargetTag admin + 1, pm)),
	      bangStrat' s0 (updateAdmin (s',lastTag,0) admin))
	 | (Source om', s') =>
             (Source (TagO (lastTag,om')), 
              bangStrat'' s0 admin s')
	end
      | _ => raise BangError)
and bangStrat'' s0 admin s' = strategy
    (fn Source (TagP (i,pm)) =>
	if i = #lastSourceTag admin + 1 
        then case s' % Source pm of
	     (Source om, s'') =>
	         (Source (TagO (i,om)), 
		  bangStrat'' s0 (incSourceTag admin) s'')
	   | (Target pm', s'') =>
		 (Target (TagP (#lastTargetTag admin + 1, pm')),
		  bangStrat' s0 (updateAdmin (s'',i,1) admin))
	else raise BangError
      | _ => raise BangError)

in
fun BANG s =  bangStrat' s initAdmin
end ;


(* Dereliction, of type !X -> X. (The same strategy works for any X.) *)

exception DerelictError ;

local 
fun derStrat' i = strategy
    (fn Target om =>
	(Source (TagO (i,om)), derStrat'' (i+1))
      | _ => raise ParityError)
and derStrat'' i = strategy
    (fn Source (TagP (j,pm)) =>
	if j=i then (Target pm, derStrat' i) 
	else raise DerelictError
      | _ => raise DerelictError)
in
val DERELICT = derStrat' 0
end ;


(* To deal with reusable objects not of form !Y (e.g. !X x !X'), we need
   a bit of machinery for manipulating "embedded bangs". This is given in
   the following structure, used for both comultiplication and contraction. *)

structure Context = struct

datatype context = 
     TrivC | ProdC of label * context | InjC of label * context ;

exception ContextError of context ;
exception NoBangErrorO of omove ;
exception NoBangErrorP of pmove ;

fun applyCtxtO TrivC om = om
  | applyCtxtO (ProdC (c,ctxt)) om = ProdO (c, applyCtxtO ctxt om)
  | applyCtxtO (InjC (c,ctxt)) om  = InjO (c, applyCtxtO ctxt om) ;

fun applyCtxtP TrivC pm = pm
  | applyCtxtP (ProdC (c,ctxt)) pm = ProdP (c, applyCtxtP ctxt pm)
  | applyCtxtP (InjC (c,ctxt)) pm  = InjP (c, applyCtxtP ctxt pm) ;

fun extractTagO (TagO (i,om)) = (om,i,TrivC)
  | extractTagO (ProdO (c,om)) = 
    (case extractTagO om of (om',i,ctxt) => (om',i,ProdC (c,ctxt)))
  | extractTagO (InjO (c,om)) =
    (case extractTagO om of (om',i,ctxt) => (om',i,InjC (c,ctxt))) 
  | extractTagO om = raise NoBangErrorO om ;

fun extractTagP (TagP (i,pm)) = (pm,i,TrivC)
  | extractTagP (ProdP (c,pm)) = 
    (case extractTagP pm of (pm',i,ctxt) => (pm',i,ProdC (c,ctxt)))
  | extractTagP (InjP (c,pm)) =
    (case extractTagP pm of (pm',i,ctxt) => (pm',i,InjC (c,ctxt))) 
  | extractTagP pm = raise NoBangErrorP pm ;

end (* structure Context *) ;


(* Comultiplication for an arbitrary reusable X (needn't be of form !Y).
   A morphism from X to !X. *)

local

open Context ;
type comultOmoveData = 
     {om : omove, source : index, ctxt : context,
      targetInner : index, targetOuter : index} ;
type comultPmoveData = 
     {pm : pmove, source : index, ctxt : context,
      targetInner : index, targetOuter : index} ;
type comultHistory = (comultOmoveData * comultPmoveData) list ;

exception ComultError ;

type comultAdmin = 
     {history : comultHistory, 
      lastSourceTags : (context * index) list,
      lastTargetOuter : index} ;
     (* may be more bookkeeping than we need - revise later *)

fun updateComultAdmin (omd,pm,i,j,k,c) (admin:comultAdmin) : comultAdmin =
    {history = (omd, {pm=pm, source=k, ctxt=c, 
		      targetInner=i, targetOuter=j}) :: #history admin,
     lastSourceTags = updateAlist (c,k) (#lastSourceTags admin), 
     lastTargetOuter = j} ;

fun nextSourceTag c (admin:comultAdmin) =
    case optLookup (#lastSourceTags admin) c of
	NONE => 1 | SOME k => k+1 ;

fun chasePointer j (admin:comultAdmin) = 
    (case List.drop (#history admin, #lastTargetOuter admin - j) of
	 [] => raise ComultError
       | (omd,pmd)::rest =>
	 if #targetOuter pmd = j then pmd else raise IndexError j) ;

local
fun chasePointers' (iOpt,c) j (history:comultHistory) len =
    (case List.drop (history, len-j) of
	 [] => NONE
       | (omd,pmd)::rest =>
	 if #targetOuter pmd <> j then raise IndexError j
	 else if #ctxt pmd = c
              andalso (case iOpt of 
		NONE => true | SOME i => i= #targetInner pmd)
         then SOME pmd
         else chasePointers' (iOpt,c) (#targetOuter omd) rest (j-1))
fun chasePointers ic' j (admin:comultAdmin) =
    chasePointers' ic' j (#history admin) (#lastTargetOuter admin)
in
fun matchingPointer (i,c) j (admin:comultAdmin) = 
    if i=0 then 0 
    else case chasePointers (SOME i,c) j admin of
	     SOME pmd => #source pmd
	   | NONE => raise ComultError
fun lastInner c j (admin:comultAdmin) =
    case chasePointers (NONE,c) j admin of
             SOME pmd => #targetInner pmd
	   | NONE => 0
end ;

fun comult' (admin:comultAdmin) = strategy
    (fn Target (TagO (j,om')) =>
	(case extractTagO om' of (om,i,ctxt) =>
         case (matchingPointer (i,ctxt) j admin) of k =>
            (Source (applyCtxtO ctxt (TagO (k,om))),
	     comult'' admin 
		 {om=om, source=k, ctxt=ctxt, 
		  targetInner=i, targetOuter=j}))
      | _ => raise ComultError)
and comult'' admin (omd as {targetOuter=j, ctxt=ctxt, ...}) = strategy
    (fn Source pm' =>
	(case extractTagP pm' of (pm,k,ctxt') =>
	      if k <> nextSourceTag ctxt admin then raise IndexError k
	      else if ctxt' <> ctxt then raise ComultError
	      else (case #lastTargetOuter admin + 1 of j' =>
		let val i = (* if j=0 then 1 else *)
			    lastInner ctxt j admin + 1   (* How's this? *)
		in (Target (TagP (j',applyCtxtP ctxt (TagP (i,pm)))),
	            comult' (updateComultAdmin (omd,pm,i,j',k,ctxt) admin))
		end))
      | _ => raise ComultError)
in

val COMULT = comult' {history=[], lastSourceTags=[], lastTargetOuter=0} 
end ;


(* Contraction for an arbitrary reusable X (needn't be of form !Y). 
   A morphism from X to an arbitrary component-indexed power of X. 
   The same strategy works for any X and any choice of component names. *)

local

open Context ;
type address = label * context ;
type omoveData = 
     {om :omove, source :index, addr :address, target :index}
type pmoveData = 
     {pm :pmove, source :index, addr :address, target :index}
type history = (omoveData * pmoveData) list ; 
     (* moves in OP pairs, most recent first *)

type admin = {history : history, 
	      lastSourceTags : (context * index) list, 
	      lastTargetTags : (address * index) list} ;

exception AddressError of address ;
exception ContractionError of Momove ;

val initAdmin : admin =
    {history = [], 
     lastSourceTags = [],
     lastTargetTags = [] } ; 
    (* source and target tags are added dynamically as required *)

fun nextSourceTag c (a:admin) = 
    case optLookup (#lastSourceTags a) c of 
	NONE => 1 | SOME n => n+1 ;

fun nextTargetTag s (a:admin) = 
    case optLookup (#lastTargetTags a) s of
	NONE => 1 | SOME n => n+1 ;

fun updateAdmin (omd,pm,c,j,i) (a:admin) =
    {history = (omd, {pm=pm, source=j, addr=c, target=i})
	       :: #history a,
     lastSourceTags = updateAlist (#2 c,j) (#lastSourceTags a),
     lastTargetTags = updateAlist (c,i) (#lastTargetTags a)} 
    : admin ;

fun getSourceTag (a:admin) c 0 = 0
  | getSourceTag (a:admin) c i = 
    case List.find (fn (_,pmd) => #target pmd = i andalso #addr pmd = c) 
		   (#history a)
     of SOME (_,pmd) => #source pmd 
      | NONE => raise IndexError i ;

fun contraction' admin = strategy
    (fn Target (ProdO (c,om)) =>
        (case extractTagO om of (om',i,ctxt) =>
	 case getSourceTag admin (c,ctxt) i of j =>
	     (Source (applyCtxtO ctxt (TagO (j,om'))),
	      contraction'' admin 
		  {om=om', source=j, addr=(c,ctxt), target=i}))
      | om => raise ContractionError om)
and contraction'' admin 
	          (omd as {addr = addr as (c,ctxt), ...}) = strategy
    (fn Source pm =>
        (case extractTagP pm of (pm',j,ctxt') =>
	if j <> nextSourceTag ctxt admin then raise IndexError j
        else if ctxt' <> ctxt then raise ContextError ctxt'
	else (case nextTargetTag addr admin of i =>
		  (Target (ProdP (c, applyCtxtP ctxt (TagP (i,pm')))),
		   contraction' (updateAdmin (omd,pm',addr,j,i) admin))))
      | om => raise ContractionError om)

in
val CONTRACT = contraction' initAdmin 
end ;


(* "Lifting" (i.e. weak coproducts) *)

local
exception LiftError
fun UnMainO (MainO om) = om | UnMainO _ = raise LiftError
fun UnMainP (MainP pm) = pm | UnMainP _ = raise LiftError
fun lift' s = 
    RELABEL {relabP = UnMainP, relabO = MainO}
    THEN s 
    THEN RELABEL {relabP = MainP, relabO = UnMainO}
in

fun LIFT (G : ground -> ground * morphism) = strategy 
    (fn Target (BasicO x) =>
        (Source (BasicO blank), strategy
            (fn Source (BasicP x) =>
	        let val (y,s) = G x in
                    (Target (BasicP y), lift' s)
                end
	      | _ => raise LiftError))
      | _ => raise LiftError)

fun UNIT g = strategy
    (fn Target _ => 
	(Target (BasicP g), 
	 RELABEL {relabP = MainP, relabO = UnMainO})
      | _ => raise LiftError)

val FLATTEN = strategy  (* L(G) -> G for *any* G, not nec lifted *)
    (fn Target om =>
        (Source (BasicO blank),
         strategy (fn Source (BasicP _) =>
            (Source (MainO om), 
             strategy (fn Source (MainP pm) =>
	         (Target pm, ID +> {relabP = UnMainP, relabO = MainO})
			| _ => raise LiftError))
		    | _ => raise LiftError))
      | _ => raise LiftError)
                        
fun STRENGTH lab ytrans = strategy
    (fn Target (BasicO x) =>
        (Source (ProdO (lab, BasicO x)),
         strategy (fn Source (ProdP (l, BasicP y)) =>
             if l=lab then
		 (Target (BasicP (ytrans y)),
                  RELABEL {relabO = fn (MainO (om as ProdO (l,o'))) =>
                               if l=lab then ProdO (l, MainO o') else om
                                     | _ => raise LiftError,
                           relabP = fn (pm as ProdP (l,p)) =>
                               if l=lab then MainP (ProdP (l, UnMainP p))
                               else MainP pm
				     | _ => raise LiftError})
             else raise LiftError
		    | _ => raise LiftError))
      | _ => raise LiftError)

end (* lifting stuff *) ;


(* Fixpoints *)

local fun FIX' om = 
    (CONTRACT THEN 
     PRODUCT [("1",DERELICT), ("2",strategy FIX')] THEN
     UNCURRY ("1","2") ID)
    % om 
in
val FIX = strategy FIX' 
end ;

exception ZeroError 
val ZERO = strategy (fn _ => raise ZeroError) ;


end (* structure CBN_Games *) ;
