(* Top level of Lingay implementation.
   Integrates lexer, parser, typechecker and evaluator to yield
   a top-level read-eval loop, which can be invoked in batch or 
   interactive mode. *)


structure LingayLrVals = LingayLrValsFun(structure Token = LrParser.Token)
structure LingayLex = LingayLexFun(structure Tokens = LingayLrVals.Tokens)
structure LingayParser =
              Join (structure ParserData = LingayLrVals.ParserData
                    structure Lex=LingayLex
                    structure LrParser=LrParser)

structure Lingay = struct

(*
structure Absyn = Absyn
structure Typechecker = Typechecker
structure RunExp = RunExp
(* reinstate these once we have civilized signatures for them *)
*)

val expDebug = ref (NONE: (Types.errpos * Types.ty) Absyn.exp option)
val envDebug = ref ([]: (string * Types.ty * Typechecker.var_safe * Typechecker.var_newsafe * Types.ty Absyn.exp) list)

fun invoke n lexstream =
    let fun print_error (s,i:int,_) =
            TextIO.output (TextIO.stdOut,
                          "Error, line " ^ (Int.toString i) ^ ", " ^ s ^ "\n")
    in LingayParser.parse (n,lexstream,print_error,())
    end

fun lintest pos tenv ty = if (Types.is_reusable_ty pos tenv ty) then RunExp.Ord else RunExp.Lin

fun makeRunExp tenv linenv e = 
    RunExp.makeRexp (Types.componentsOf tenv) (Types.classDataOf tenv) 
		    (lintest (0,0) tenv) (Types.trivGroundFor tenv) 
		    linenv e

exception ConformError ;

local open Pervasives ; open Types ;
      val lookup = errorLookup (fn _ => ConformError)
in
fun conformToType tenv (NamedTy t) v = conformToType tenv (lookup tenv t) v
  | conformToType _ _ blankv = blankv
  | conformToType _ BoolTy (boolv b) = boolv b
  | conformToType _ IntTy (intv b) = intv b
  | conformToType _ ForestTy (foldv f) = foldv f
  | conformToType _ _ (prodv [("#1",_), ("#2",_)]) = blankv
  | conformToType _ _ (prodv [("#2",_), ("#1",_)]) = blankv
  | conformToType tenv (ProdTy (tys,ext)) (prodv clist) =
              (* add alphasort? *)
        prodv (map' (List.filter (fn (l,_) => List.exists (fn (l',_) => l'=l) tys) clist)
		    (fn (l,x) => (l, conformToType tenv (lookup tys l) x)))
  | conformToType tenv (SumTy tys) (injv (l,x)) =
        injv (l, conformToType tenv (lookup tys l) x)
  | conformToType tenv (BangTy ty) v = conformToType tenv ty v
  | conformToType tenv (ty as RecTy _) (foldv x) = 
        foldv (conformToType tenv (recbody (0,0) tenv ty) x)
  | conformToType _ _ _ = raise ConformError
end ;

fun parseStream lookahead s prompt initialStuff = 
    let open Types ; open Absyn ; open Typechecker ;
        val () = LingayLex.UserDeclarations.pos := 0
                                   
	fun extend env x t tag = (x,t,tag)::env
	fun mklexer () = LingayParser.makeLexer (Aux.inputLine s)
        val lexer =  mklexer()
        val dummyEOF = LingayLrVals.Tokens.EOF(0,0)
                       
        fun errorlines (left, right) = if left = right then ("Error on line "^(Int.toString (left+1))^": ")
                       else ("Error on lines "^(Int.toString (left+1))^"--"^(Int.toString (right+1))^": ")

	fun tryDelete (set, x) = VarSet.delete(set, x) handle NotFound => set

        fun do_valdecl tenv env linenv vars runenv bindvar safety exp = 
        let val (exp', vars') = typecheck tenv (map (fn (x,t,s1,s2,e)=>(x,t,s1)) env) exp 
            val () = debugPrint "Plain typecheck completed.\n"
            val (pos,ty) = exp_top exp'
            val linstatus = lintest pos tenv ty
	    val () = lincheck pos [vars,vars']
	    val () = debugPrint "Variables checked, now inferring newsafety information.\n"
            val exp'' = newcheck tenv (map (fn (x,t,s1,s2,e) => (x,(s1,s2))) env) exp' 
            val newsafeStatus = h_newsafetag exp'' 
	    val () = debugPrint "Done, now safety checking.\n"
            val () = safecheck tenv (map (fn (x,t,s1,s2,e) => (x,(s1,s2))) env) exp'' safety
            val () = TextIO.print ("    "^bindvar^" : "^ (tytostring ty)^" "^(tagtostring safety)
				   ^ (case newsafeStatus of Newunsafevar => " [not newsafe]" | _ => "")  ^"\n") 
            val () = (expDebug := SOME exp')

            val run_exp = makeRunExp tenv linenv (Absyn.exp_map snd exp')
            val run_exp' = RunExp.annWithFVs run_exp
            val runenv' =
		let val (expval,runenv') = Semantics.updateEnv runenv run_exp' (bindvar, linstatus)
		    val () = TextIO.print ("    "^bindvar^" = "^ 
					   Pervasives.toString (conformToType tenv ty expval) ^ "\n")
		in runenv' end
		handle Semantics.NotImplemented s => 
		       (TextIO.print("Runtime warning: "^s^" not yet implemented\n") ; runenv)
		     | ConformError =>
                       (TextIO.print("Runtime error: expression may involve unsupported language feature.\n") ; runenv)
					      
            val varsafety = case safety of 
				Ssyn.Safe => Safevar
			      | Ssyn.Metasafe => Unsafevar (* Metasafevar *)
			      | Ssyn.Argsafe => Argsafevar
			      | _ => Unsafevar
	    val varnewsafety = case newsafeStatus of 
				   Halfnewsafevar => Newsafevar | s => s
            val env1 = (bindvar, ty, varsafety, varnewsafety, (exp_map (fn (p,t)=>t) exp')) :: env
            val () = (envDebug := env1)
        in (env1, (bindvar,linstatus)::linenv, tryDelete (VarSet.union(vars,vars'), bindvar), runenv') end
        handle Untypeable (pos, s) => (TextIO.print ((errorlines pos)^"Untypable: "^s^"\n"); (env,linenv,vars,runenv))
             | Unsafe (pos, s) => (TextIO.print((errorlines pos)^"Untypable (safety violation): "^s^"\n"); (env,linenv,vars,runenv))
             | Unimplemented n => (TextIO.print("Internal error: unimplemented feature: "^(Int.toString n)^"\n"); (env,linenv,vars,runenv))
             | Impossible s => (TextIO.print("Internal error: "^s^"\n"); (env,linenv,vars,runenv))
             | Overflow => (TextIO.print("Runtime error: Integer overflow.\n"); (env,linenv,vars,runenv))
			   (* handle other runtime errors? *)

	fun do_typedecl pos tenv tname ty = 
	    (case status [] pos (resolvety pos tenv ty) of SOME stat => (TextIO.print("type "^tname^" = "^(tytostring ty)^"\n");  extendTy tenv tname ty)
				   | NONE => (TextIO.print("Bad type in type declaration ("^tname^"): "^(tytostring ty)^"\n"); tenv))
	    handle Impossible s => (TextIO.print("Internal error: "^s^"\n"); tenv)
		 | Unimplemented _  => (TextIO.print("Unimplemented."); tenv)
                 | Unsafe _  => (TextIO.print("Internal error (unsafe)."); tenv)
                 | Untypeable (pos,s) => (TextIO.print((errorlines pos)^s^"\n"); tenv)

        fun loop lexer (stuff as (tenv, (env,_,_,_))) =
            let val () = TextIO.print prompt
		val (result,lexer) = invoke lookahead lexer
                val (nextToken,lexer) = LingayParser.Stream.get lexer
                (* for purpose of named superclass type annotation (inefficient): *)
		val _ = (Absyn.currStaticEnvStore := map (fn (x,t,_,_,_) => (x,t)) env)
                val _ = (Absyn.currTypeEnvStore := tenv) 
                val translated = Absyn.translateTopdec result
                val _ = map Absyn.display translated 
                val allStuff as (tenv',(env',linenv',vars',runenv')) = 
                    List.foldl
                        (fn (dc, (tenv,(env,linenv,vars,runenv))) =>
                            case dc of
                              Exp exp => (tenv, do_valdecl tenv env linenv vars runenv "it" None exp)
                            | Val (x,exp,safety) => (tenv, do_valdecl tenv env linenv vars runenv x safety exp)
                            | Type (tname, ty) =>  (do_typedecl (0,0) tenv tname ty, (env,linenv,vars,runenv))
                        (*  | Empty => (tenv, (env,linenv,vars,runenv))  *)
                        )
                        stuff translated
            in
                if LingayParser.sameToken(nextToken,dummyEOF) then allStuff
                else loop lexer allStuff
            end
            handle  LingayParser.ParseError => 
                    let val () = TextIO.print("Parse error.\n")
                        val (nextToken,lexer) = LingayParser.Stream.get lexer 
			val lexer = mklexer()
                    in loop lexer stuff
		    end
                  | Absyn.Unfinished s =>
                    let val () = TextIO.print("Syntax not yet supported: "^s^"\n")
                        val (nextToken,lexer) = LingayParser.Stream.get lexer 
			val lexer = mklexer()
                    in loop lexer stuff
		    end
                  | Absyn.UnknownSuperclass x =>
                    let val () = TextIO.print("Unknown named superclass: "^x^"\n")
                        val (nextToken,lexer) = LingayParser.Stream.get lexer 
			val lexer = mklexer()
                    in loop lexer stuff
		    end
    in loop lexer initialStuff
    end

val empty_env = Typechecker.empty_env ;
val emptyStuff = (empty_env, (empty_env,empty_env,VarSet.empty,Semantics.initialEnv))

fun startRet initialStuff = parseStream 0 TextIO.stdIn "Lingay> " initialStuff ;
fun start () = (startRet emptyStuff; ())
fun processFileRet f = parseStream 15 (TextIO.openIn f) (if !Typechecker.debug then "." else "") emptyStuff
fun loadFile f = let val fileStuff = processFileRet f in
                      startRet fileStuff ; ()
                 end

val GC_messages = SMLofNJ.Internals.GC.messages ;
val _ = GC_messages false ;

(* Debugging utilities *)
fun get_exp() = valOf (!expDebug) ;
fun run_exp e = makeRunExp [] [] (Absyn.exp_map (fn (x,y)=>y) e) ;
fun annotate re = RunExp.annWithFVs re ;
val interpret = Semantics.interpret ;
val initVal = let open Pervasives in
              prodv [("ord",prodv []),("lin",prodv [])] 
	      end 

end (* structure Lingay *) ;
