(* File:    RunExp.sml
   Author:  John Longley
   Date:    1 August 2008, 00:04 (started)
*)

(* Datatype of runtime expressions and associated bureaucracy *)


structure RunExp = struct

(* Runtime expressions. Typecasts etc stripped out for runtime purposes. 
   A few other annotations added: e.g. label names in Merge and Extend,
   and linearity status of all variable occurrences. *)

open Pervasives

type varname = string   
type label = string
datatype linopt = Lin | Ord
type var = varname * linopt
datatype const = Bool of bool | Int of int

(* Runtime expressions with 'a annotations at each node *)

datatype 'a Rexp = 
    Var of 'a * var
  | Const of 'a * const
  | Product of 'a * (label * 'a Rexp) list
  | Merge of 'a * 'a Rexp * label list * 'a Rexp * label list  
             (* left and right component labels listed *)
  | Split of 'a * 'a Rexp * label list *  (* all labels listed *)
	     (label * var) list * var * 'a Rexp
  | Inject of 'a * label * 'a Rexp
  | Match of 'a * 'a Rexp * (label * var * 'a Rexp) list
  | Prom of 'a * 'a Rexp
  | Der of 'a * 'a Rexp
  | Fold of 'a * 'a Rexp
  | Unfold of 'a * 'a Rexp
  | Oncefn of 'a * var * 'a Rexp
  | App of 'a * 'a Rexp * 'a Rexp
  | Eq of 'a * 'a Rexp * 'a Rexp
  | Infix of 'a * 'a Rexp * infixop * 'a Rexp
  | If of 'a * 'a Rexp * 'a Rexp * 'a Rexp
  | Rec of 'a * var * groundv * 'a Rexp
  | Catchcont of 'a * var * 'a Rexp
  | Force of 'a * 'a Rexp
  | Triv of 'a * linopt
  | Extend of 'a * linopt * 
              label list * label list * label list * label list *
              (* parent fields, parent methods, new fields, new methods *)
	      'a Rexp * 'a Rexp * 'a Rexp
  | New of 'a * linopt * label list * label list * 'a Rexp  
	      (* all fields and methods listed *)
  | Make of 'a * linopt * label list * label list * 'a Rexp * 'a Rexp  
	      (* ditto *)



(* Constructing a (bare) runtime expression from a 
   type-annotated compile-time expression *)

exception RunExpError

fun lookup x l = 
    let val printDebug = fn _ => ()   (* or TextIO.print *) 
        val () = printDebug ("Looking up "^x^" in "^
				String.concat (map fst l)^"... ")
        val y = errorLookup (fn _ => RunExpError) l x
        val () = printDebug ("Done\n")
    in y end

fun coerceLo lo = if lo then Lin else Ord
fun infixOf Ssyn.IntPlus = IntPlus
  | infixOf Ssyn.IntMinus = IntMinus
  | infixOf Ssyn.IntLessThan = IntLess

fun isIn v varlist = List.exists (fn (v',_) => v'=v) varlist

fun makeRexp getComps getClassData getStatus getTriv linEnv e = 
    let val mk' = makeRexp getComps getClassData getStatus getTriv
        val mk  = mk' linEnv
        fun labs e = map fst (getComps (Absyn.exp_top e)) 
	fun addStatus varlist comptypes =
	    map' varlist (fn (l,v) => 
		 (l, (v, getStatus (lookup l comptypes))))
	fun remainderStatus varlist comptypes =
	    coerceLo (exists' comptypes (fn (v,ty) => 
			not (isIn v varlist) andalso getStatus ty = Lin))
    in case e of 
	Absyn.Var (_,v) => Var ((), (v, lookup v linEnv))
      | Absyn.Const (_,Absyn.Bool b) => Const ((), Bool b)
      | Absyn.Const (_,Absyn.Int i) => Const ((), Int i)
      | Absyn.Product (_,complist) => 
	  Product ((), map (fn (l,e) => (l, mk e)) complist)
      | Absyn.Merge (_, e1, e2) => 
	  Merge ((), mk e1, labs e1, mk e2, labs e2)
      | Absyn.Split (_, e1, varlist, v, e2) =>
          let val comptypes = getComps (Absyn.exp_top e1) 
              val comptypes' = addStatus varlist comptypes
              val v' = (v, remainderStatus varlist comptypes)
          in  Split ((), mk e1, map fst comptypes, comptypes',
                     v', mk' (map snd comptypes' @ v'::linEnv) e2)
          end
      | Absyn.Inject (_, l, e) => Inject ((), l, mk e)
      | Absyn.Match (_, e, caselist) => 
          let val comptypes = getComps (Absyn.exp_top e) in
	      Match ((), mk e, map' caselist 
	                (fn Absyn.MatchCase (_,l,v,e') => 
                            let val v' = (v, getStatus (lookup l comptypes))
			    in (l, v', mk' (v'::linEnv) e')
			    end))
          end
      | Absyn.Prom (_, e) => Prom ((), mk e)
      | Absyn.Der (_, e) => Der ((), mk e)
      | Absyn.Force (_, e) => Force ((), mk e)
      | Absyn.Fold (_, ty, e) => Fold ((), mk e)
      | Absyn.Unfold (_, e) => Unfold ((), mk e)
      | Absyn.Fun (_, v, ty, e) => 
	  let val v' = (v, getStatus ty) in
	      Oncefn ((), v', mk' (v'::linEnv) e)
          end
      | Absyn.App (_, e1, e2) => App ((), mk e1, mk e2)
      | Absyn.Eq (_, e1, e2) => Eq ((), mk e1, mk e2)
      | Absyn.Infix (_, iop, e1, e2) => 
	  Infix ((), mk e1, infixOf iop, mk e2)
      | Absyn.If (_, e1, e2, e3) => If ((), mk e1, mk e2, mk e3)
      | Absyn.Rec (_, v, ty, e) => 
	  let val v' = (v, getStatus ty) in
	      Rec ((), v', getTriv ty, mk' (v'::linEnv) e)
          end
      | Absyn.Catchcont (_, v, ty, e) => 
	  let val v' = (v, getStatus ty) in
	      Catchcont ((), v', mk' (v'::linEnv) e)
          end
      | Absyn.Cast (_, e, _) => mk e
      | Absyn.Triv (_, lo) => Triv ((), coerceLo lo)
      | Absyn.Extend (top,lo,e1,e2,e3) =>
          let val (flds, meths, _) = getClassData (Absyn.exp_top e1) ;
              val (flds',meths', _) = getClassData top ;
              val flds'' = Pervasives.diff flds' flds
              val meths'' = Pervasives.diff meths' meths
	  in  Extend ((), coerceLo lo, flds, meths, flds'', meths'', 
		      mk e1, mk e2, mk e3)
          end
      | Absyn.Constr (_,e) =>
          let val (flds, meths, lo) = getClassData (Absyn.exp_top e) in
	      New ((), coerceLo lo, flds, meths, mk e)
	  end
      | Absyn.Make (_,e1,e2) =>
          let val (flds, meths, lo) = getClassData (Absyn.exp_top e1) in
	      Make ((), coerceLo lo, flds, meths, mk e1, mk e2)
	  end
      | Absyn.Catchcopy _ => raise RunExpError
    end ;
 

fun annOf (Var (a,_)) = a
  | annOf (Const (a,_)) = a
  | annOf (Product (a,_)) = a
  | annOf (Merge (a,_,_,_,_)) = a
  | annOf (Split (a,_,_,_,_,_)) = a
  | annOf (Inject (a,_,_)) = a
  | annOf (Match (a,_,_)) = a
  | annOf (Prom (a,_)) = a
  | annOf (Der (a,_)) = a
  | annOf (Fold (a,_)) = a
  | annOf (Unfold (a,_)) = a
  | annOf (Oncefn (a,_,_)) = a
  | annOf (App (a,_,_)) = a
  | annOf (Eq (a,_,_)) = a
  | annOf (Infix (a,_,_,_)) = a
  | annOf (If (a,_,_,_)) = a
  | annOf (Rec (a,_,_,_)) = a
  | annOf (Catchcont (a,_,_)) = a
  | annOf (Force (a,_)) = a
  | annOf (Triv (a,_)) = a
  | annOf (Extend (a,_,_,_,_,_,_,_,_)) = a
  | annOf (New (a,_,_,_,_)) = a
  | annOf (Make (a,_,_,_,_,_)) = a

type runexp = (varname list * varname list) Rexp

fun combineAnn (lins,ords) =
    List.@ (map' lins (fn n => (n,Lin)),
	    map' ords (fn n => (n,Ord)))

local
fun mergeAnns (lin1,ord1) (lin2,ord2) =
    (List.@ (lin1,lin2), Pervasives.merge ord1 ord2) 
fun mergeAllAnns [] = ([],[])
  | mergeAllAnns (FVs::rest) = mergeAnns FVs (mergeAllAnns rest)
fun removeVar (name, Lin) (lin,ord) = (remove name lin, ord)
  | removeVar (name, Ord) (lin,ord) = (lin, remove name ord)
fun removeVars [] FVs = FVs
  | removeVars (v::rest) FVs = removeVars rest (removeVar v FVs)
in
fun annWithFVs (Var (_,v as (name,lin))) =
    Var (case lin of Lin => ([name],[]) | Ord => ([],[name]), v)
  | annWithFVs (Const (_,c)) = Const (([],[]),c)
  | annWithFVs (Product (_,complist)) = 
    let val complist' = map' complist (fn (l,e) => (l, annWithFVs e))
    in Product (mergeAllAnns (map (annOf o snd) complist'), complist')
    end
  | annWithFVs (Merge (_,e1,labs1,e2,labs2)) = 
    let val e1' = annWithFVs e1
        val e2' = annWithFVs e2
    in Merge (mergeAnns (annOf e1') (annOf e2'), e1', labs1, e2', labs2)
    end
  | annWithFVs (Split (_,e1,labs1,labvars,v,e2)) = 
    let val e1' = annWithFVs e1
        val e2' = annWithFVs e2
    in Split (removeVars (v::map snd labvars) 
	          (mergeAnns (annOf e1') (annOf e2')),
              e1', labs1, labvars, v, e2')
    end 
  | annWithFVs (Inject (_,lab,e)) = 
    let val e' = annWithFVs e
    in Inject (annOf e',lab,e')
    end
  | annWithFVs (Match (_,e1,caselist)) = 
    let val e1' = annWithFVs e1
        val caselist' = map' caselist (fn (l,v,e) => (l,v, annWithFVs e))
    in Match (mergeAnns (annOf e1')
              (mergeAllAnns (map' caselist' 
		   (fn (l,v,e') => removeVar v (annOf e')))),
              e1', caselist')
    end
  | annWithFVs (Prom (_,e)) = 
    let val e' = annWithFVs e
    in Prom (annOf e',e')
    end
  | annWithFVs (Der (_,e)) = 
    let val e' = annWithFVs e
    in Der (annOf e',e')
    end
  | annWithFVs (Fold (_,e)) = 
    let val e' = annWithFVs e
    in Fold (annOf e',e')
    end
  | annWithFVs (Unfold (_,e)) = 
    let val e' = annWithFVs e
    in Unfold (annOf e',e')
    end
  | annWithFVs (Oncefn (_,v,e)) = 
    let val e' = annWithFVs e
    in Oncefn (removeVar v (annOf e'),v,e')
    end
  | annWithFVs (App (_,e1,e2)) = 
    let val e1' = annWithFVs e1
        val e2' = annWithFVs e2
    in App (mergeAnns (annOf e1') (annOf e2'), e1', e2')
    end
  | annWithFVs (Eq (_,e1,e2)) = 
    let val e1' = annWithFVs e1
        val e2' = annWithFVs e2
    in Eq (mergeAnns (annOf e1') (annOf e2'), e1', e2')
    end
  | annWithFVs (Infix (_,e1,iop,e2)) = 
    let val e1' = annWithFVs e1
        val e2' = annWithFVs e2
    in Infix (mergeAnns (annOf e1') (annOf e2'), e1', iop, e2')
    end
  | annWithFVs (If (_,e1,e2,e3)) = 
    let val e1' = annWithFVs e1
        val e2' = annWithFVs e2
        val e3' = annWithFVs e3
    in If (mergeAllAnns [annOf e1', annOf e2', annOf e3'], e1', e2', e3')
    end
  | annWithFVs (Rec (_,v,x0,e)) = 
    let val e' = annWithFVs e
    in Rec (removeVar v (annOf e'), v, x0, e')
    end
  | annWithFVs (Catchcont (_,v,e)) = 
    let val e' = annWithFVs e
    in Catchcont (removeVar v (annOf e'), v, e')
    end
  | annWithFVs (Force (_,e)) = 
    let val e' = annWithFVs e
    in Force (annOf e',e')
    end
  | annWithFVs (Triv (_,lo)) = Triv (([],[]), lo)
  | annWithFVs (Extend (_,lo,fields,meths,fields',meths',e1,e2,e3)) =
    let val e1' = annWithFVs e1
        val e2' = annWithFVs e2
        val e3' = annWithFVs e3
    in Extend (mergeAllAnns [annOf e1', annOf e2', annOf e3'],
	       lo, fields, meths, fields', meths', e1', e2', e3')
    end 
  | annWithFVs (New (_,lo,fields,methods,e)) = 
    let val e' = annWithFVs e
    in New (annOf e',lo,fields,methods,e')
    end
  | annWithFVs (Make (_,lo,fields,methods,e1,e2)) =
    let val e1' = annWithFVs e1
        val e2' = annWithFVs e2
    in Make (mergeAnns (annOf e1') (annOf e2'), lo,fields,methods,e1',e2')
    end
end

end (* structure RunExp *) ;
