(* File:    Semantics.sml
   Author:  John Longley
   Date:    1 August 2008
*)

(* Core of Lingay runtime based on game semantics.
   Denotations of runtime expressions are strategies defined
   compositionally by recursion on syntactic structure. *)

signature SEMANTICS = sig
    type varname = string
    type runexp
    type denotation
    type ground
    type environment
    val interpret  : runexp -> denotation
    val initialEnv : environment
    val updateEnv  : environment -> runexp ->
		     (varname * RunExp.linopt) -> (ground * environment)
    exception NotImplemented of string
end ;

functor Semantics (
    structure CBV : CBV_GAMES 
    ) : SEMANTICS = struct

open Pervasives
open RunExp 
open CBV

type denotation = vmorph
infix THEN THENL


(* Some derived morphisms (simple category theory!) *)

fun ZERO x = GROUND (fn _ => x)

infix THENL ;  (* co-Kleisli composition *)
fun F THENL F' = F THEN LIFT F' THEN FLATTEN ;

fun PROMOTE F = COMULT THEN BANG F ;

local 
    fun SMASH_ALL' [] = ID
      | SMASH_ALL' (l::rest) = 
        LIFT (STRENGTH l) THEN FLATTEN THEN SMASH_ALL' rest
in 
fun SMASH_ALL (l::rest) = STRENGTH l THEN SMASH_ALL' rest
  | SMASH_ALL [] = ZERO (prodv[]) THEN UNIT
end

val TRIV = BANG (CURRY ("","#") (GROUND (fn _ => prodv []) THEN UNIT))

fun COMULT_ALL labs = PRODUCT (map (fn lab => (lab,COMULT)) labs)


(* Pairs and triples *)

fun PAIR (F1,F2) = PRODUCT [("#1",F1), ("#2",F2)]

fun pairv (x1,x2) = prodv [("#1",x1), ("#2",x2)]

fun TRIPLE (F1,F2,F3) = PRODUCT [("#1",F1), ("#2",F2), ("#3",F3)]

val STRENGTH_1 = STRENGTH "#1"

val STRENGTH_2 = STRENGTH "#2"

val SMASH_PAIR = SMASH_ALL ["#1","#2"]

val SMASH_TRIPLE = SMASH_ALL ["#1","#2","#3"]

val DISTRIB_PAIR = DISTRIB "#2" "#1"

val CURRY_PAIR = CURRY ("#1","#2")

val APPLY_PAIR = UNCURRY ("#1","#2") ID

fun APPLY_TO F = 
    PROD_INJ "#1" ["#2"] 
    THEN PAIR (ID,F) THEN APPLY_PAIR

val FLATTEN_RHS = CURRY_PAIR (APPLY_PAIR THEN FLATTEN) 

val COND_TRIPLE = COND ("#1","#2","#3")


(* Structural manipulations of environments *)

val trivlinordv = prodv [("lin", prodv[]), ("ord", prodv[])]

fun STRUCT_COMPS' compvarlist =
    let val lablist = map fst compvarlist 
        val linSplit = concatmap' compvarlist 
	    (fn (lab,linvars) => map' linvars 
		(fn name => ([name], [name,lab])))
        val distrib = concatmap' lablist 
	    (fn lab => [ ([lab,"lin"], ["lin",lab]),
                         ([lab,"ord"], ["ord",lab]) ])
        val matrix = prodv (map' lablist (fn lab => (lab, trivlinordv)))
    in  PRODUCT [("ord", (CONTRACT lablist)), 
		 ("lin", (PROJ linSplit))] 
	THEN (PROJ' distrib matrix)
    end 

fun STRUCT_COMPS complist =
    STRUCT_COMPS' (mapsnd (fn e => fst (annOf e)) complist)

fun STRUCT_PAIR (e1,e2) = 
    STRUCT_COMPS [("#1",e1), ("#2",e2)] 

fun STRUCT_TRIPLE (e1,e2,e3) = 
    STRUCT_COMPS [("#1",e1), ("#2",e2), ("#3",e3)] 

fun STRUCT_PAIR' (e,e0) =
    let val (linvars, linvars0) = (fst (annOf e), fst (annOf e0))
    in STRUCT_COMPS' [("#1",diff linvars linvars0), ("#2",linvars0)]
    end 

(* Bureaucratic component relabellings *)

fun linlab Lin = "lin"
  | linlab Ord = "ord"

fun MERGE_PRODUCTS labs1 labs2 =
    PROJ (List.@ (map' (diff labs1 labs2) (fn l => ([l,"#1"], [l])),
		  map' labs2 (fn l => ([l,"#2"], [l]))))

fun SPLIT_PROJ labs1 vars2 labvars (restvar as (restname,restlo)) =
    PROJ' ((map' labvars 
		(fn (l,(name,lo)) => ([l,"#1"], [name,linlab lo]))) @
          (map' (diff labs1 (map fst labvars))
	        (fn l => ([l, "#1"], [l,restname,linlab restlo]))) @
          (map' (diff vars2 (restvar :: map snd labvars))
                (fn (name,lo) => ([name,linlab lo,"#2"], 
				  [name,linlab lo])))) 
          (* matrix: *)
          (case restlo of
	       RunExp.Lin => prodv [("lin", prodv [(restname,prodv[])]), 
			     ("ord", prodv [])]
	     | RunExp.Ord => prodv [("lin", prodv []), 
			     ("ord", prodv [(restname,prodv[])])])

fun INCORP_PROJ vars (var as (name,lo)) =
    PROJ' ((["#2"],[name,linlab lo]) ::
           map' vars (fn (name',lo') => 
		 	 ([name',linlab lo',"#1"], [name',linlab lo'])))
          trivlinordv  (* matrix *)

val EXTEND_PROJ = PROJ'
    [(["#1","#1"], ["superMethods","ord"]),
     (["#2","#1"], ["superConstr","ord"]),
     (["#2"],      ["methodsImpl","ord"]),
     (["#3"],      ["constrImpl","ord"])] 
   trivlinordv ;

val REASSOC = PROJ 
    [(["#1","#1"], ["#1"]),  
     (["#2","#1"], ["#1","#2"]),  
     (["#2"],      ["#2","#2"])] ;

exception NotImplemented of string
exception RuntimeTypeError

fun unfoldv (foldv x) = x
  | unfoldv _ = raise RuntimeTypeError

fun trivVal methods = prodv (map' methods (fn m => (m,blankv)))

fun toGround (Bool b) = boolv b
  | toGround (Int i) = intv i


(* Ordinary class extension, implemented via a Lingay program.
   (Cf. the function "extendOp" in the file "theory.er"). *)

fun Fn (v,e)     = Prom ((), Oncefn ((),v,e)) ;
fun IApp (e,e')  = App  ((), Der ((),e), e') ;
fun Dot (e,ms,m) = Split ((), e, ms, [(m,("x",Ord))], ("y",Ord),
			  Var ((),("x",Ord))) ;

val superMethods = Var ((),("superMethods",Ord))
val superConstr  = Var ((),("superConstr",Ord))
val methodsImpl  = Var ((),("methodsImpl",Ord))
val constrImpl   = Var ((),("constrImpl",Ord))
val this         = Var ((),("this",Ord))
val selfrw       = Var ((),("selfrw",Ord))
val superrw      = Var ((),("superrw",Ord))
val superrw'     = Var ((),("superrw'",Ord))
fun mkvar f      = Var ((),(f,Ord))

fun extendProgram0 oldflds oldmths newflds newmths = 
    let val allflds = oldflds @ newflds
        val allmths = oldmths @ newmths
    in
    Product ((), 
       [("#1", Fn (("this",Ord),
         IApp 
            (IApp 
                (methodsImpl,
                 Product ((),
                 map' oldmths
                 (fn m => (m, Fn (("selfrw",Ord),
                  Split ((), 
                      selfrw,
                      allflds,
		      map' newflds (fn f => (f,(f,Ord))),
                      ("superrw",Ord),
                      IApp (Dot 
                         (IApp 
                             (superMethods,
                              Product ((),
				  map' oldmths
                                  (fn m' => (m',
				      Fn (("superrw'",Ord),
                                          IApp 
                                             (Dot (this, allmths, m'),
					      Merge ((),
                                                 superrw',
                                                 oldflds,
                                                 Product ((),
                                                    map' newflds
                                                    (fn f => (f,mkvar f))),
                                                 newflds))))))) ,
                          oldmths, m),
                          superrw))))))),
             this))) ,
        ("#2", IApp (constrImpl, superConstr))])
    end ;

fun extendProgram oldflds oldmths newflds newmths = 
    RunExp.annWithFVs (extendProgram0 oldflds oldmths newflds newmths) ;


(* Now the interesting bit:
   Denotational semantics of programs, given compositionally
   as a function from syntax trees to morphisms in CBV game model. *)

fun interpret (Var (_,(name,lo))) =
      PROJ [([name,linlab lo], [])] THEN UNIT

  | interpret (Const (_,c)) =
      ZERO (toGround c) THEN UNIT

  | interpret (Product (_,complist)) =
      STRUCT_COMPS complist 
      THEN (PRODUCT (mapsnd interpret complist))
      THEN (SMASH_ALL (map fst complist))

  | interpret (Merge (_,e1,labs1,e2,labs2)) =
      STRUCT_PAIR (e1,e2) 
      THEN (PAIR (interpret e1, interpret e2)) 
      THEN SMASH_PAIR 
      THEN (LIFT (MERGE_PRODUCTS labs1 labs2))

  | interpret (Split (_,e1,labs,labvars,v,e2)) =
      STRUCT_PAIR (e1,e2)  
      THEN PAIR (interpret e1, ID) 
      THEN STRENGTH_1
      THENL (SPLIT_PROJ labs (combineAnn (annOf e2)) labvars v
             THEN interpret e2)

  | interpret (Inject (_,lab,e)) = 
      interpret e THEN LIFT (TAG lab)

  | interpret (e as Match (a,e0,caselist)) = 
      STRUCT_PAIR' (e,e0)
      THEN PAIR (ID, interpret e0)
      THEN STRENGTH_2
      THENL (DISTRIB_PAIR
             THEN CASES (map' caselist (fn (lab,x,e') =>
                  (lab, INCORP_PROJ (combineAnn a) x 
                        THEN interpret e'))))

  | interpret (Prom (_,e)) =  (* NB: e always a value here *)
      LPROMOTE (PROJ [(["ord"],["ord"])] 
                THEN interpret e) 

  | interpret (Der (_,e)) =
      interpret e THEN LIFT DERELICT

  | interpret (Fold (_,e)) = 
      interpret e THEN LIFT (RELABEL foldv)

  | interpret (Unfold (_,e)) = 
      interpret e THEN LIFT (RELABEL unfoldv)
				       
  | interpret (Oncefn (a,v,e)) =
      CURRY_PAIR (INCORP_PROJ (combineAnn a) v 
	          THEN interpret e)
      THEN UNIT

  | interpret (App (_,e1,e2)) =
      STRUCT_PAIR (e1,e2) 
      THEN PAIR (interpret e1, interpret e2) 
      THEN SMASH_PAIR 
      THENL APPLY_PAIR 

  | interpret (Eq (_,e1,e2)) =
      STRUCT_PAIR (e1,e2) 
      THEN PAIR (interpret e1, interpret e2) 
      THEN SMASH_PAIR 
      THEN LIFT (GROUND GroundEq)

  | interpret (Infix (_,e1,iop,e2)) =
      (STRUCT_PAIR (e1,e2)) 
      THEN (PAIR (interpret e1, interpret e2)) 
      THEN SMASH_PAIR 
      THEN (LIFT (GROUND (GroundInfix iop)))

  | interpret (If (_,e1,e2,e3)) =
      STRUCT_TRIPLE (e1,e2,e3)
      THEN TRIPLE (interpret e1, interpret e2, interpret e3)
      THEN STRENGTH_1
      THENL COND_TRIPLE

  | interpret (Rec (a,v,x0,e)) =
      PROJ [(["ord"],["ord"])] 
      THEN COMULT 
      THEN BANG (CURRY_PAIR
         ((INCORP_PROJ (combineAnn a) v) THEN interpret e THEN FLATTEN))
      THEN FIX x0 THEN UNIT

  | interpret (Catchcont (a,v,e)) = 
      CURRY_PAIR (INCORP_PROJ (combineAnn a) v 
	          THEN interpret e)
      THEN CATCHCONT

  | interpret (Force (_,e)) = 
      interpret e THENL FORCE

  (* Class machinery. We take the denotation of a non-linear class type 
          classimpl state, object, argtype end
     to be the denotation of
          { #1 : \tau_self -> \tau_self , 
            #2 : argtype -> state }
     where if object = { m_0 : !\tau_0, ... } then
           \tau_self = { m_0 : RWall(state) -> \tau_0, ... }
     Recall that the denotation of "->" involves \bot, -o and !. *)

  | interpret (Triv (_,lo)) =
      ZERO (pairv (blankv, blankv))  
      THEN PAIR (TRIV, TRIV) THEN UNIT
(*
  (* Easy case of Extend: *)
  | interpret (Extend (_, Ord, [], [], [], [], Triv(_,Ord), e1, e2)) = 
      STRUCT_PAIR (e1,e2) 
      THEN PAIR (interpret e1 
                 THENL (DERELICT THEN APPLY_TO (ZERO (prodv []))) ,
		 interpret e2 
                 THENL (DERELICT THEN APPLY_TO TRIV))
      THEN SMASH_PAIR 
*)
  | interpret (Extend (_, Ord, oldflds, oldmths, newflds, newmths,
                       e0, e1, e2)) =
      STRUCT_TRIPLE (e0, e1, e2)
      THEN TRIPLE (interpret e0, interpret e1, interpret e2)
      THEN SMASH_TRIPLE
      THENL (EXTEND_PROJ THEN
             interpret (extendProgram oldflds oldmths newflds newmths))

  | interpret (Extend (_, Lin, _,_,_,_,_,_,_)) =
    raise NotImplemented "linear class extension"

  | interpret (New (_,Ord,fields,methods,e)) =
      (interpret e)
      THENL (PAIR ((BANG FLATTEN_RHS 
		    THEN FIX (trivVal methods)) 
                   THEN (THREAD methods), 
		   ID)
             THEN  (PROMOTE (CURRY_PAIR 
                     (REASSOC
                      THEN PAIR (DERELICT ,
                                 (PAIR (DERELICT, ID) 
                                 THEN APPLY_PAIR 
                                 THEN LIFT (COMULT_ALL fields))) 
                      THEN STRENGTH_2
                      THENL APPLY_PAIR)))
             THEN UNIT)

  | interpret (New (_,Lin,fields,methods,e)) = 
    raise NotImplemented "linear class instantiation"

  | interpret (Make _) =
    raise NotImplemented "make-instantiation"

  (* To be added: Extend, New and Make for linear classes *)


(* Top-level dynamic environments *)

type varname = string
type environment = denotation * varname list * varname list
		   (* list linear and ordinary vars available for use *)

val initialEnv : environment = (ZERO trivlinordv, [], [])

fun memoize F = INCLUDE (CBV.` (F, blankv))

fun updateEnv (envSem, envLinvars, envOrdvars) e (bindvar,lo) =
    let val eLinvars = fst (annOf e)
        val otherLinvars = diff envLinvars eLinvars
        val env2 = envSem THEN STRUCT_COMPS' 
			       [("#1", eLinvars), ("#2", otherLinvars)]
        val beforeEval = env2 THEN PAIR (interpret e, ID)
	val (v,afterEval) = CBV.evaluate "#1" beforeEval
        val incorpBind = 
	    (PROJ' ((["#1"], [bindvar, linlab lo]) ::
                    map' (remove bindvar envLinvars)
                         (fn v => ([v,"lin","#2"],[v,"lin"])) @
                    map' (remove bindvar envOrdvars)
                         (fn v => ([v,"ord","#2"],[v,"ord"])))
		   trivlinordv)
    in (v, (memoize (afterEval THEN incorpBind), 
            if lo=RunExp.Lin then insert bindvar envLinvars else envLinvars,
	    if lo=RunExp.Ord then insert bindvar envOrdvars else envOrdvars))
    end 

end (* functor Semantics *) ;
